# pyright: reportUnknownMemberType=false
"""Rich content classes."""

from __future__ import annotations

import base64
import io
from typing import TYPE_CHECKING
import urllib.request
import xml.etree.ElementTree as ET

import rico._config
import rico._html


if TYPE_CHECKING:
    from typing import Any, Literal


try:
    import altair as alt
    import vl_convert as vlc
except ImportError:
    alt = None

try:
    import markdown
except ImportError:
    markdown = None

try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = None

try:
    import seaborn.objects as so
except ImportError:
    so = None


class ContentBase:
    """A base content definition.

    Creates a container element on init.
    Defines serialization method and string representation.

    Attributes:
        container (Element): The container element.
    """
    container: ET.Element

    def __init__(self, class_: str | None = None):
        """Initialize base content.

        Args:
            class_: The container element's class attribute.
        """
        attrib = {"class": class_} if class_ is not None else {}
        self.container = ET.Element("div", attrib=attrib)

    def serialize(
        self,
        indent: bool | None = None,
        space: str | None = None,
        strip: bool | None = None,
    ) -> str:
        """Serialize the object to string in HTML format.

        Indent the object if `indent_space` is not None.

        Args:
            indent: If True, indent the element.
            space: The whitespace for indentation.
            strip: If True, strip unnecessary whitespace.

        Returns:
            The serialized object.
        """
        return rico._html.serialize_html(
            self.container, indent=indent, space=space, strip=strip)

    def __str__(self) -> str:
        """Serialize the object to string in HTML format."""
        return self.serialize()


class Tag(ContentBase):
    """A Tag content definition.

    Creates a content element using tag parameters and appends it to the container.
    """

    def __init__(
        self,
        tag: str,
        text: str | None = None,
        tail: str | None = None,
        attrib: dict[str, Any] = {},
        class_: str | None = None,
        **extra: Any,
    ):
        """Initialize content using tag parameters.

        Args:
            tag: The content element's tag.
            text: Text before first subelement.
            tail: Text after the end tag.
            attrib: The tag's attributes.
            extra: Extra attributes.
            class_: The container class attribute.
        """
        super().__init__(class_)
        element = ET.Element(tag, attrib={**attrib, **extra})
        element.text = text
        element.tail = tail
        self.container.append(element)


class Text(ContentBase):
    """A Text content definition.

    Creates a content element from a text and appends it to the container.
    """
    def __init__(
        self,
        obj: Any,
        mono: bool = False,
        wrap: bool = False,
        class_: str | None = None,
    ):
        """Initialize content from a text.

        The default tag is <p> unless the text contains a line break.
        Then the <pre> tag is used.

        The `mono` and `wrap` parameters rely on Bootstrap CSS.

        Args:
            obj: The text. If it's not an instance of str, then it's converted to str.
            mono: If True then "font-monospace" class is assigned to the text element.
            wrap: If True then "text-wrap" class is assigned to the text element.
            class_: The container class attribute.
        """
        super().__init__(class_)

        if not isinstance(obj, str):
            obj = str(obj)

        text_class = " ".join([
            cl
            for cond, cl in [(mono, "font-monospace"), (wrap, "text-wrap")]
            if cond
        ])

        tag = "pre" if "\n" in obj else "p"
        attrib = {"class": text_class} if text_class else {}
        element = ET.Element(tag, attrib=attrib)
        element.text = obj
        self.container.append(element)


class Code(ContentBase):
    """A Code content definition.

    Creates content elements from a code and appends them to the container.
    """
    def __init__(self, text: str, class_: str | None = None):
        """Initialize content from a code.

        Args:
            text: The code.
            class_: The container class attribute.
        """
        super().__init__(class_)
        pre = ET.Element("pre")
        code = ET.Element("code")
        code.text = text
        pre.append(code)
        self.container.append(pre)


class HTML(ContentBase):
    """An HTML content definition.

    Creates content elements from an HTML text and appends them to the container.
    """
    def __init__(
        self,
        text: str,
        strip_dataframe_borders: bool = False,
        class_: str | None = None,
    ):
        """Initialize content from an HTML text.

        Args:
            text: The HTML text.
            strip_dataframe_borders: Delete borders attributes from dataframes.
            class_: The container class attribute.
        """
        super().__init__(class_)

        for element in rico._html.parse_html(text):
            self.container.append(element)

        if strip_dataframe_borders:
            for table in self.container.iterfind(
                './/table[@class="dataframe"][@border]',
            ):
                del table.attrib["border"]


class Markdown(ContentBase):
    """A Markdown content definition.

    Creates content elements from a markdown text and appends them to the container.
    """
    def __init__(
        self,
        text: str,
        class_: str | None = None,
        **kwargs: Any,
    ):
        """Initialize content from a markdown text.

        Args:
            text: The markdown text.
            class_: The container class attribute.
            **kwargs: Keyword arguments passed to the `markdown.markdown` function.

        Raises:
            ImportError: The markdown package is not installed.
        """
        if markdown is None:
            raise ImportError("The markdown package is not installed.")

        content = HTML(markdown.markdown(text, **kwargs), class_=class_)
        self.container = content.container


class Image(ContentBase):
    """An Image content definition.

    Creates content elements using an image data and appends them to the container.
    """
    def __init__(
        self,
        data: bytes | str,
        format: str | None = None,  # noqa: A002
        class_: str | None = None,
    ):
        """Initialize content using image data.

        Args:
            data: The image data.
            format: The image format.
            class_: The container class attribute.
        """
        super().__init__(class_)

        if format is None:
            format = rico._config.get_config("image_format")  # noqa: A001

        if format == "svg":
            if isinstance(data, bytes):
                data = data.decode()

            for element in rico._html.parse_html(data):
                self.container.append(element)
        else:
            if isinstance(data, str):
                data = data.encode()
            encoded_image = base64.b64encode(data).decode()

            element = ET.Element(
                "img",
                attrib={"src": f"data:image/{format};base64,{encoded_image}"},
            )
            self.container.append(element)


class Chart(ContentBase):
    """An Chart content definition.

    Creates content elements from a chart object and appends them to the container.

    The supported chart types are the following:
    - Altair Chart,
    - Pyplot Axes and Figure,
    - Seaborn Plot (seaborn.objects interface).
    """
    def __init__(
        self,
        obj: Any,
        format: Literal["svg", "png"] | None = None,  # noqa: A002
        class_: str | None = None,
        **kwargs: Any,
    ):
        """Initialize content from a chart object.

        Args:
            obj: The chart object.
            format: The image format.
            class_: The container class attribute.
            **kwargs: Keyword arguments passed to a function
                which converts object to an image.

        Raises:
            TypeError: Chart type is not supported
                or required extra package is not installed.
        """
        if format is None:
            format = rico._config.get_config("chart_format")  # noqa: A001

        if plt is not None and isinstance(obj, plt.Axes):
            obj = obj.figure

        if plt is not None and isinstance(obj, plt.Figure):  # type: ignore
            stream = io.StringIO() if format == "svg" else io.BytesIO()
            obj.savefig(stream, format=format, **kwargs)
            image = stream.getvalue()
        elif so is not None and isinstance(obj, so.Plot):
            stream = io.StringIO() if format == "svg" else io.BytesIO()
            obj.save(stream, format=format, **kwargs)
            image = stream.getvalue()
        elif alt is not None and isinstance(obj, alt.Chart):
            convert = vlc.vegalite_to_svg if format == "svg" else vlc.vegalite_to_png  # type: ignore  # noqa: E501
            image = convert(  # type: ignore
                obj.to_json(),
                vl_version="_".join(alt.SCHEMA_VERSION.split(".")[:2]),
                **kwargs,
            )
        else:
            error_msg = (
                f"Chart type {type(obj)} is not supported "
                "or required extra package is not installed."
            )
            raise TypeError(error_msg)

        content = Image(data=image, format=format, class_=class_)  # type: ignore
        self.container = content.container


class Obj(ContentBase):
    """An arbitrary content definition.

    Creates content elements from arbitrary objects and appends them to the container.

    Automatically determines the content type.
    """
    def __init__(self, *objects: Any, class_: str | None = None):
        """Initialize content from arbitrary objects.

        Args:
            *objects: The objects which are used to create a content.
            class_: The container class attribute.
        """
        super().__init__(class_=class_)
        for obj in objects:
            if (
                alt is not None and isinstance(obj, alt.Chart) or
                plt is not None and isinstance(obj, plt.Axes | plt.Figure) or  # type: ignore  # noqa: E501
                so is not None and isinstance(obj, so.Plot)
            ):
                content = Chart(obj)
            elif hasattr(obj, "_repr_html_") and callable(obj._repr_html_):
                content = HTML(obj._repr_html_(), strip_dataframe_borders=True)
            else:
                content = Text(obj)

            for element in content.container:
                self.container.append(element)


class Script(ContentBase):
    """A script definition.

    Attributes:
        script (Element): The script element.
        footer (bool): Defines whether the script should be placed at a document footer,
            aftert all other content.
    """
    script: ET.Element
    footer: bool = False

    def __init__(
        self,
        text: str | None = None,
        src: str | None = None,
        inline: bool = False,
        defer: bool = False,
        attrib: dict[str, Any] = {},
        **extra: Any,
    ):
        """Initialize script.

        Either `text` or `src` should be used.

        Args:
            text: The script text.
            src: The script source link.
            inline: If True, load script inline, downdload it from source link
                and use it as a text.
            defer: The script "defer" attribute. For inline script defines whether
                it should be placed at a document footer, aftert all other content.
            attrib: The script attributes.
            **extra: Extra attributes.

        Raises:
            ValueError: Both text and src are not None.
            ValueError: Both text and src are None.
        """
        if text is not None and src is not None:
            raise ValueError("Either `text` or `src` should be None.")

        if text is None and src is None:
            raise ValueError("Either `text` or `src` should be not None.")

        if inline and src is not None:
            with urllib.request.urlopen(src) as response:
                text = response.read().decode()
            src = None

        if src is not None:
            attrib = {"src": src, **attrib}
            if defer:
                attrib = {"defer": True, **attrib}
        else:
            self.footer = defer

        self.script = ET.Element("script", {**attrib, **extra})
        self.script.text = text
        self.container = self.script


class Style(ContentBase):
    """A stylesheet definition.

    Attributes:
        style (Element): The style element.
    """
    style: ET.Element

    def __init__(
        self,
        text: str | None = None,
        src: str | None = None,
        inline: bool = False,
        attrib: dict[str, Any] = {},
        **extra: Any,
    ):
        """Initialize stylesheet.

        Either `text` or `src` should be used.

        Args:
            text: The stylesheet text.
            src: The stylesheet source link.
            inline: If True, load stylesheet inline, downdload it from source link
                and use it as a text.
            attrib: The stylesheet attributes.
            **extra: Extra attributes.

        Raises:
            ValueError: Both text and src are not None.
            ValueError: Both text and src are None.
        """
        if text is not None and src is not None:
            raise ValueError("Either `text` or `src` should be None.")

        if text is None and src is None:
            raise ValueError("Either `text` or `src` should be not None.")

        if inline and src is not None:
            with urllib.request.urlopen(src) as response:
                text = response.read().decode()
            src = None

        if src is not None:
            tag = "link"
            attrib = {"src": src, **attrib}
            if "rel" not in attrib:
                attrib = {**attrib, "rel": "stylesheet"}
        else:
            tag = "style"

        self.style = ET.Element(tag, {**attrib, **extra})
        self.style.text = text
        self.container = self.style



# pyright: reportPrivateUsage=false
from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

import rico._config


if TYPE_CHECKING:
    from collections.abc import Generator
    from typing import Any


@pytest.fixture
def reset_config() -> Generator[None, Any, None]:  # noqa: PT004
    old_config = rico._config._global_config.copy()

    try:
        yield
    finally:
        rico._config._global_config = old_config


@pytest.mark.usefixtures("reset_config")
def test_get_config():
    config = rico._config.get_config()
    assert config == rico._config._global_config
    config["indent_html"] = not config["indent_html"]
    assert config != rico._config._global_config

    assert (
        rico._config.get_config("indent_html") ==
        rico._config._global_config["indent_html"]
    )


@pytest.mark.usefixtures("reset_config")
def test_set_config():
    rico._config.set_config(indent_html=True)
    assert rico._config._global_config["indent_html"] is True

    rico._config.set_config(indent_html=False)
    assert rico._config._global_config["indent_html"] is False


@pytest.mark.usefixtures("reset_config")
def test_config_context():
    old_indent_html = rico._config._global_config["indent_html"]

    with rico._config.config_context(indent_html=not old_indent_html):
        assert rico._config._global_config["indent_html"] is not old_indent_html

    assert rico._config._global_config["indent_html"] is old_indent_html



# ruff: noqa: ARG001
"""Global config."""

from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from collections.abc import Generator
    from typing import Any, Literal


_global_config = {
    "indent_html": False,
    "indent_space": "  ",
    "strip_html": False,
    "chart_format": "svg",
    "image_format": "svg",
}


def get_config(param: str | None = None) -> Any:
    """Get global configuration.

    Args:
        param: Paramater name.

    Returns:
        Parameter value if `param` is not None, or dict with all parameters otherwise.
    """
    config = _global_config.copy()
    if param is not None:
        return config[param]
    return config


def set_config(
    indent_html: bool | None = None,
    indent_space: str | None = None,
    strip_html: bool | None = None,
    chart_format: Literal["svg", "png"] | None = None,
    image_format: str | None = None,
) -> None:
    """Set global configuration.

    Args:
        indent_html: Indent HTML elements in serialization methods.
        indent_space: Default indent space.
        strip_html: Strip HTML elements in serialization methods.
        chart_format: Default chart image format.
        image_format: Default image format.
    """
    for param, value in locals().items():
        if value is not None:
            _global_config[param] = value


@contextlib.contextmanager
def config_context(
    indent_html: bool | None = None,
    indent_space: str | None = None,
    strip_html: bool | None = None,
    chart_format: Literal["svg", "png"] | None = None,
    image_format: str | None = None,
) -> Generator[None, Any, None]:
    """Context manager for configuration.

    Args:
        indent_html: Indent HTML elements in serialization methods.
        indent_space: Default indent space.
        strip_html: Strip HTML elements in serialization methods.
        chart_format: Default chart image format.
        image_format: Default image format.
    """
    new_config = locals()
    old_config = get_config()
    set_config(**new_config)

    try:
        yield
    finally:
        set_config(**old_config)



"""HTML parser and serializer."""

from __future__ import annotations

import html.parser
import xml.etree.ElementTree as ET

import rico._config


TAGS_EMPTY = {
    "area", "base", "basefont", "br", "col", "embed", "frame", "hr", "img",
    "input", "isindex", "link", "meta", "param", "path", "source", "track", "wbr",
}

TAGS_INLINE = {
    "a", "abbr", "b", "bdi", "bdo", "br", "cite", "code", "data", "dfn", "em",
    "i", "kbd", "mark", "q", "rp", "rt", "ruby", "s", "samp", "small", "span",
    "strong", "sub", "sup", "time", "u", "var", "wbr",
}

TAGS_NOT_ESCAPED = {"script", "style"}
TAGS_PREFORMATTED = {"pre"}


class HTMLParser(html.parser.HTMLParser):
    """Simple HTML parser. Returns a list of instances of ET.Element on close().

    Assigns None values to boolean attributes.

    Ignores comments, doctype declaration and processing instructions.
    Converts attribute names to lower case, even for SVG.
    """

    def __init__(self):
        super().__init__()
        self._root = "root"
        self._builder = ET.TreeBuilder()
        self._builder.start(self._root, {})

    def handle_starttag(
        self,
        tag: str,
        attrs: list[tuple[str, str | None]],
    ) -> None:
        self._builder.start(tag, dict(attrs))

    def handle_endtag(self, tag: str) -> None:
        self._builder.end(tag)

    def handle_data(self, data: str) -> None:
        self._builder.data(data)

    def close(self) -> list[ET.Element]:
        super().close()
        self._builder.end(self._root)
        return list(self._builder.close())


def parse_html(data: str) -> list[ET.Element]:
    """Parse an HTML document from a string.

    Assign None values to boolean attributes.

    Ignore comments, doctype declaration and processing instructions.
    Convert attribute names to lower case, even for SVG.

    Args:
        data: The HTML data.

    Returns:
        The parsed HTML document.
    """
    parser = HTMLParser()
    parser.feed(data)
    return parser.close()


def indent_html(
    element: ET.Element,
    space: str = "  ",
    level: int = 0,
) -> ET.Element:
    """Indent an HTML element.

    Tnsert newlines and indentation space after elements.
    Create a new element instead of updating inplace.
    Do not indent elements inside <pre> tag.

    Args:
        element: The element to indent.
        space: The whitespace to insert for each indentation level.
        level: The initial indentation level. Should always be 0.

    Returns:
        The indented HTML element.
    """
    if element.tag.lower() in TAGS_PREFORMATTED or not len(element):
        return element

    indented_element = ET.Element(element.tag, attrib=element.attrib)
    indented_element.text = element.text
    indented_element.tail = element.tail

    if not element.text or not element.text.strip():
        indented_element.text = "\n" + space * (level + 1)

    for child in element:
        indented_child = indent_html(child, space=space, level=level + 1)

        if not indented_child.tail or not indented_child.tail.strip():
            indented_child.tail = "\n" + space * (level + 1)

        indented_element.append(indented_child)

    if not indented_child.tail or not indented_child.tail.strip():  # type: ignore
        indented_child.tail = "\n" + space * level  # type: ignore

    return indented_element


def strip_html(element: ET.Element) -> ET.Element:
    """Strip an HTML element.

    Remove unnecessary whitespaces from the element by strippping elements'
    text and tail.
    Do not strip elements inside <pre> tag or inside inline tags.

    Args:
        element: The element to strip.

    Returns:
        The stripped HTML element.
    """
    stripped_element = ET.Element(element.tag, attrib=element.attrib)
    stripped_element.text = element.text
    stripped_element.tail = element.tail

    if element.tag.lower() in TAGS_INLINE | TAGS_PREFORMATTED:
        for child in element:
            stripped_element.append(child)
    else:
        for child in element:
            stripped_element.append(strip_html(child))

        if stripped_element.text:
            if len(stripped_element) == 0 and (
                not stripped_element.tail or
                not stripped_element.tail.strip()
            ):
                stripped_element.text = stripped_element.text.strip()
            else:
                stripped_element.text = stripped_element.text.lstrip()

    if stripped_element.tail:
        if len(stripped_element) == 0 and not stripped_element.text:
            stripped_element.tail = stripped_element.tail.strip()
        else:
            stripped_element.tail = stripped_element.tail.rstrip()

    return stripped_element


def _escape_cdata(text: str) -> str:
    """Copy of xml.etree.ElementTree._escape_cdata."""
    if "&" in text:
        text = text.replace("&", "&amp;")
    if "<" in text:
        text = text.replace("<", "&lt;")
    if ">" in text:
        text = text.replace(">", "&gt;")
    return text

def _escape_attrib_html(text: str) -> str:
    """Copy of xml.etree.ElementTree._escape_attrib_html."""
    if "&" in text:
        text = text.replace("&", "&amp;")
    if ">" in text:
        text = text.replace(">", "&gt;")
    if '"' in text:
        text = text.replace('"', "&quot;")
    return text

def serialize_html(
    element: ET.Element,
    indent: bool | None = None,
    space: str | None = None,
    strip: bool | None = None,
) -> str:
    """Serialize an HTML element to a string.

    Serialize attributes with None values as boolean.

    Args:
        element: The HTML element.
        indent: If True, indent the element.
        space: The whitespace for indentation.
        strip: If True, strip unnecessary whitespace.

    Returns:
        The serialized HTML element.
    """
    global_config = rico._config.get_config()
    if indent is None:
        indent = global_config["indent_html"]
    if space is None:
        space = global_config["indent_space"]
    if strip is None:
        strip = global_config["strip_html"]

    if strip:
        element = strip_html(element)
    if indent:
        element = indent_html(element, space=space)  # type: ignore

    attrib = "".join(
        f' {k}="{_escape_attrib_html(v)}"'
        if v is not None and not isinstance(v, bool)  # type: ignore
        # Serialize attributes with None values as boolean.
        else f" {k}"
        for k, v in element.items()
        if not isinstance(v, bool) or v
    )

    opening_tag = f"<{element.tag}{attrib}>"
    ltag = element.tag.lower()

    if element.text is not None:
        text = element.text if ltag in TAGS_NOT_ESCAPED else _escape_cdata(element.text)
    else:
        text = ""

    children = "".join(serialize_html(e) for e in element)
    closing_tag = f"</{element.tag}>" if ltag not in TAGS_EMPTY else ""
    tail = _escape_cdata(element.tail) if element.tail is not None else ""

    return opening_tag + text + children + closing_tag + tail



from __future__ import annotations

import textwrap
import xml.etree.ElementTree as ET

import pytest

import rico._html


def elem_to_string(elem: ET.Element | list[ET.Element], sep: str = "") -> str:
    if isinstance(elem, list):
        return sep.join(elem_to_string(e) for e in elem)
    return ET.tostring(elem, encoding="unicode", method="html")


def test_html_parser_two_elements():
    text = "<p>Hello</p><p>world</p>"
    parser = rico._html.HTMLParser()
    parser.feed(text)
    elements = parser.close()

    assert elem_to_string(elements) == text
    assert isinstance(elements, list)
    assert len(elements) == 2

    p0 = elements[0]
    assert isinstance(p0, ET.Element)
    assert p0.tag == "p"
    assert p0.attrib == {}
    assert p0.text == "Hello"
    assert p0.tail is None
    assert len(p0) == 0

    p1 = elements[1]
    assert isinstance(p1, ET.Element)
    assert p1.tag == "p"
    assert p1.attrib == {}
    assert p1.text == "world"
    assert p1.tail is None
    assert len(p1) == 0


def test_html_parser_nested_tags():
    text = "<div><p>Hello <strong>world</strong>!</p></div>"
    parser = rico._html.HTMLParser()
    parser.feed(text)
    elements = parser.close()

    assert elem_to_string(elements) == text
    assert isinstance(elements, list)
    assert len(elements) == 1

    div = elements[0]
    assert isinstance(div, ET.Element)
    assert div.tag == "div"
    assert div.attrib == {}
    assert div.text is None
    assert div.tail is None
    assert len(div) == 1

    p = list(div)[0]
    assert isinstance(p, ET.Element)
    assert p.tag == "p"
    assert p.attrib == {}
    assert p.text == "Hello "
    assert p.tail is None
    assert len(p) == 1

    strong = list(p)[0]
    assert isinstance(strong, ET.Element)
    assert strong.tag == "strong"
    assert strong.attrib == {}
    assert strong.text == "world"
    assert strong.tail == "!"
    assert len(strong) == 0


def test_html_parser_attributes():
    script_src = "https://cdn.jsdelivr.net/npm/bootstrap@5.3.0-alpha3/dist/js/bootstrap.bundle.min.js"
    text = f"<script defer src='{script_src}' crossorigin='anonymous'></script>"
    parser = rico._html.HTMLParser()
    parser.feed(text)
    elements = parser.close()

    assert isinstance(elements, list)
    assert len(elements) == 1

    script = elements[0]
    assert isinstance(script, ET.Element)
    assert script.tag == "script"
    assert script.attrib == {
        "defer": None,
        "src": script_src,
        "crossorigin": "anonymous",
    }
    assert script.text is None
    assert script.tail is None
    assert len(script) == 0


def test_html_parser_svg():
    text = (
        '<svg xmlns="http://www.w3.org/2000/svg" '
        'xmlns:xlink="http://www.w3.org/1999/xlink" '
        'width="16" height="16" fill="currentColor" class="bi bi-dash">'
        '<path d="M4 8a.5.5 0 0 1 .5-.5h7a.5.5 0 0 1 0 1h-7A.5.5 0 0 1 4 8z"/>'
        "</svg>"
    )
    parser = rico._html.HTMLParser()
    parser.feed(text)
    elements = parser.close()

    assert isinstance(elements, list)
    assert len(elements) == 1

    svg = elements[0]
    assert isinstance(svg, ET.Element)
    assert svg.tag == "svg"
    assert svg.attrib == {
        "xmlns": "http://www.w3.org/2000/svg",
        "xmlns:xlink": "http://www.w3.org/1999/xlink",
        "width": "16",
        "height": "16",
        "fill": "currentColor",
        "class": "bi bi-dash",
    }
    assert svg.text is None
    assert svg.tail is None
    assert len(svg) == 1

    path = list(svg)[0]
    assert isinstance(path, ET.Element)
    assert path.tag == "path"
    assert path.attrib == {
        "d": "M4 8a.5.5 0 0 1 .5-.5h7a.5.5 0 0 1 0 1h-7A.5.5 0 0 1 4 8z",
    }
    assert path.text is None
    assert path.tail is None
    assert len(path) == 0


def test_parse_html():
    text = "<p>Hello world</p>"
    elements = rico._html.parse_html(text)

    assert elem_to_string(elements) == text
    assert isinstance(elements, list)
    assert len(elements) == 1

    p = elements[0]
    assert isinstance(p, ET.Element)
    assert p.tag == "p"
    assert p.attrib == {}
    assert p.text == "Hello world"
    assert p.tail is None
    assert len(p) == 0


@pytest.fixture
def sample_elem():
    div0 = ET.Element("div", {"class": "container"})
    div0.text = "\n"
    p0 = ET.SubElement(div0, "p")
    p0.text = " Hello "
    p0.tail = "\n"
    strong = ET.SubElement(p0, "strong")
    strong.text = " world "
    strong.tail = " ! "
    div1 = ET.SubElement(div0, "div", {"class": '>&"'})
    div1.text = "\n"
    div1.tail = "\n"
    code0 = ET.SubElement(div1, "code")
    code0.text = " should be indented "
    code0.tail = "\n"
    pre = ET.SubElement(div0, "pre")
    pre.text = "\n"
    pre.tail = "\n"
    code1 = ET.SubElement(pre, "code")
    code1.text = " should not be indented "
    code1.tail = "\n"
    p1 = ET.SubElement(div0, "p")
    p1.text = " Hello >&< "
    p1.tail = "\n"
    br = ET.SubElement(p1, "br")
    br.tail = " world again "
    return div0


def test_indent_html_default(sample_elem: ET.Element):
    expectation = textwrap.dedent("""\
        <div class="container">
          <p> Hello <strong> world </strong> ! </p>
          <div class="&gt;&amp;&quot;">
            <code> should be indented </code>
          </div>
          <pre>
        <code> should not be indented </code>
        </pre>
          <p> Hello &gt;&amp;&lt; <br> world again </p>
        </div>""")

    assert elem_to_string(rico._html.indent_html(sample_elem)) == expectation


def test_indent_html_custom_space(sample_elem: ET.Element):
    expectation = textwrap.dedent("""\
        <div class="container">
            <p> Hello <strong> world </strong> ! </p>
            <div class="&gt;&amp;&quot;">
                <code> should be indented </code>
            </div>
            <pre>
        <code> should not be indented </code>
        </pre>
            <p> Hello &gt;&amp;&lt; <br> world again </p>
        </div>""")

    assert elem_to_string(
        rico._html.indent_html(sample_elem, "    ")) == expectation


def test_strip_html(sample_elem: ET.Element):
    p = ET.Element("p")
    p.text = " Hello world again again "
    sample_elem.append(p)

    expectation = (
        '<div class="container"><p>Hello <strong> world </strong> !</p>'
        '<div class="&gt;&amp;&quot;"><code> should be indented </code></div><pre>\n'
        "<code> should not be indented </code>\n"
        "</pre><p>Hello &gt;&amp;&lt; <br>world again</p>"
        "<p>Hello world again again</p></div>"
    )

    assert elem_to_string(rico._html.strip_html(sample_elem)) == expectation


def test_serialize_html_default(sample_elem: ET.Element):
    expectation = textwrap.dedent("""\
        <div class="container">
        <p> Hello <strong> world </strong> ! </p>
        <div class="&gt;&amp;&quot;">
        <code> should be indented </code>
        </div>
        <pre>
        <code> should not be indented </code>
        </pre>
        <p> Hello &gt;&amp;&lt; <br> world again </p>
        </div>""")

    assert rico._html.serialize_html(sample_elem) == expectation


def test_serialize_html_indent(sample_elem: ET.Element):
    expectation = textwrap.dedent("""\
        <div class="container">
            <p> Hello <strong> world </strong> ! </p>
            <div class="&gt;&amp;&quot;">
                <code> should be indented </code>
            </div>
            <pre>
        <code> should not be indented </code>
        </pre>
            <p> Hello &gt;&amp;&lt; <br> world again </p>
        </div>""")

    assert rico._html.serialize_html(
        sample_elem, indent=True, space="    ") == expectation


def test_serialize_html_strip(sample_elem: ET.Element):
    expectation = (
        '<div class="container"><p>Hello <strong> world </strong> !</p>'
        '<div class="&gt;&amp;&quot;"><code> should be indented </code></div><pre>\n'
        "<code> should not be indented </code>\n"
        "</pre><p>Hello &gt;&amp;&lt; <br>world again</p></div>"
    )

    assert rico._html.serialize_html(sample_elem, strip=True) == expectation


def test_serialize_html_bool_attr(sample_elem: ET.Element):
    sample_elem.set("autofocus", None)  # type: ignore
    expectation = textwrap.dedent("""\
        <div class="container" autofocus>
        <p> Hello <strong> world </strong> ! </p>
        <div class="&gt;&amp;&quot;">
        <code> should be indented </code>
        </div>
        <pre>
        <code> should not be indented </code>
        </pre>
        <p> Hello &gt;&amp;&lt; <br> world again </p>
        </div>""")

    assert rico._html.serialize_html(sample_elem) == expectation
    sample_elem.set("autofocus", True)  # type: ignore
    assert rico._html.serialize_html(sample_elem) == expectation

    sample_elem.set("autofocus", False)  # type: ignore
    expectation = textwrap.dedent("""\
        <div class="container">
        <p> Hello <strong> world </strong> ! </p>
        <div class="&gt;&amp;&quot;">
        <code> should be indented </code>
        </div>
        <pre>
        <code> should not be indented </code>
        </pre>
        <p> Hello &gt;&amp;&lt; <br> world again </p>
        </div>""")
    assert rico._html.serialize_html(sample_elem) == expectation


def test_serialize_html_style():
    elem = ET.Element("style")
    elem.text = ".>&< {border: none;}"
    assert rico._html.serialize_html(elem) == "<style>.>&< {border: none;}</style>"
